/* wipemem multiboot executable - David A. Madore - Public Domain */
/* conio.c - console I/O functions */

#include <stdint.h>

#include "machine.h"
#include "conio.h"

typedef struct {
  unsigned char ch;
  uint8_t at;
} screen_line[SCREEN_COLS];

static screen_line *const screen = (screen_line *)0x000b8000UL;

static uint8_t current_attr;
static int cursor_line;
static int cursor_col;

void
select_attr (uint8_t attr)
{
  current_attr = attr;
}

static void
position_cursor (void)
{
  uint16_t pos = (cursor_line*SCREEN_COLS+cursor_col);
  outb (0x3d4, 0xf);
  outb (0x3d5, pos&0xff);
  outb (0x3d4, 0xe);
  outb (0x3d5, pos>>8);
}

void
move_cursor (int line, int col)
{
  cursor_line = line;
  cursor_col = col;
  position_cursor ();
}

void
clear_screen (void)
{
  for ( int i=0 ; i<SCREEN_LINES ; i++ )
    for ( int j=0 ; j<SCREEN_COLS ; j++ )
      {
	screen[i][j].ch = 0;
	screen[i][j].at = current_attr;
      }
  move_cursor (0, 0);
}

static void
scroll_screen (void)
{
  for ( int i=0 ; i<SCREEN_LINES-1 ; i++ )
    for ( int j=0 ; j<SCREEN_COLS ; j++ )
      {
	screen[i][j].ch = screen[i+1][j].ch;
	screen[i][j].at = screen[i+1][j].at;
      }
  for ( int j=0 ; j<SCREEN_COLS ; j++ )
    {
      screen[SCREEN_LINES-1][j].ch = 0;
      screen[SCREEN_LINES-1][j].at = current_attr;
    }
  cursor_line--;
  position_cursor ();
}

void
put_raw_ch (unsigned char ch)
{
  if ( cursor_col >= SCREEN_COLS )
    {
      cursor_col = 0;
      cursor_line++;
    }
  if ( cursor_line >= SCREEN_LINES )
    scroll_screen ();
  screen[cursor_line][cursor_col].ch = ch;
  screen[cursor_line][cursor_col].at = current_attr;
  cursor_col++;
  position_cursor ();
}

static void
cursor_nl (void)
{
  cursor_col = 0;
  cursor_line++;
  if ( cursor_line >= SCREEN_LINES )
    scroll_screen ();
  position_cursor ();
}

void
put_ch (unsigned char ch)
{
  if ( ch == '\n' )
    cursor_nl ();
  else
    put_raw_ch (ch);
}

void
put_str (const unsigned char *s)
{
  while ( *s )
    put_ch (*(s++));
}

static const unsigned int powers_of_ten[10] = {
  1U, 10U, 100U, 1000U, 10000U,
  100000U, 1000000U, 10000000U, 100000000U, 1000000000U
};
static const char hexdigits[16] = "0123456789abcdef";

void
printf (const char *fmt, ...)
{
  void *arg = &fmt + 1;
  while ( *fmt )
    {
      if ( *fmt == '%' )
	{
	  char opt_zero = 0;
	  char opt_llong = 0;
	  while ( 1 )
	    {
	      fmt++;
	      if ( *fmt == 0 )
		return;
	      else if ( *fmt == '%' )
		{
		  put_ch ('%');
		  break;
		}
	      else if ( *fmt == 's' )
		{
		  unsigned char *s = *(unsigned char **)arg;
		  arg = (unsigned char **)arg+1;
		  put_str (s);
		  break;
		}
	      else if ( *fmt == 'x' )
		{
		  if ( opt_llong ) {
		    unsigned long long v = *(unsigned long long *)arg;
		    arg = (unsigned long long *)arg+1;
		    char prt = opt_zero;
		    for ( int d=15 ; d>=0 ; d-- )
		      {
			unsigned int vd = (v>>(d*4))&0xf;
			prt = prt || vd || (d==0);
			if ( prt )
			  put_ch (hexdigits[vd]);
		      }
		  } else {
		    unsigned int v = *(unsigned int *)arg;
		    arg = (unsigned int *)arg+1;
		    char prt = opt_zero;
		    for ( int d=7 ; d>=0 ; d-- )
		      {
			unsigned int vd = (v>>(d*4))&0xf;
			prt = prt || vd || (d==0);
			if ( prt )
			  put_ch (hexdigits[vd]);
		      }
		  }
		  break;
		}
	      else if ( *fmt == 'u' )
		{
		  unsigned int v = *(unsigned int *)arg;
		  arg = (unsigned int *)arg+1;
		  char prt = opt_zero;
		  for ( int d=9 ; d>=0 ; d-- )
		    {
		      unsigned int vd = (v/powers_of_ten[d])%10;
		      prt = prt || vd || (d==0);
		      if ( prt )
			put_ch (hexdigits[vd]);
		    }
		  break;
		}
	      else if ( *fmt == 'd' )
		{
		  int v = *(int *)arg;
		  arg = (unsigned int *)arg+1;
		  if ( v < 0 )
		    {
		      put_ch ('-');
		      v = -v;
		    }
		  char prt = opt_zero;
		  for ( int d=9 ; d>=0 ; d-- )
		    {
		      unsigned int vd = (v/powers_of_ten[d])%10;
		      prt = prt || vd || (d==0);
		      if ( prt )
			put_ch (hexdigits[vd]);
		    }
		  break;
		}
	      else if ( *fmt == '0' )
		opt_zero = 1;
	      else if ( *fmt == 'L' )
		opt_llong = 1;
	    }
	}
      else
	put_ch (*fmt);
      fmt++;
    }
}
