/* wipemem multiboot executable - David A. Madore - Public Domain */

/* Known arguments are: verbose, noreboot */

#include <stdint.h>

#include "multiboot.h"
#include "machine.h"
#include "conio.h"

#define CLEAR_ATTR 0x0f
#define TEXT_ATTR 0x1f
#define ERR_ATTR 0x4f

static void
die (const char *errmsg)
{
  select_attr (ERR_ATTR);
  printf ("%s\n", errmsg);
  stop ();
}

static uint32_t
get_cr4 (void)
{
  uint32_t cr4;
  asm volatile ("mov %%cr4, %0" : "=q" (cr4));
  return cr4;
}

static void
set_cr4 (uint32_t cr4)
{
  asm volatile ("mov %0, %%cr4" : : "q" (cr4));
}

static void
set_cr3 (uint32_t cr3)
{
  asm volatile ("mov %0, %%cr3" : : "q" (cr3));
}

static uint32_t
get_cr0 (void)
{
  uint32_t cr0;
  asm volatile ("mov %%cr0, %0" : "=q" (cr0));
  return cr0;
}

static void
set_cr0 (uint32_t cr0)
{
  asm volatile ("mov %0, %%cr0" : : "q" (cr0));
}

extern volatile uint64_t pd[512];  // Page directory
extern volatile uint64_t pdpt[4];  // Page directory pointer table

static const uint64_t pd_flags   = 0x0000000000000001ULL;  // P bit
static const uint64_t page_flags = 0x0000000000000083ULL;  // P, R/W and PS bits
static const uint64_t cdwt_flags = 0x0000000000000018ULL;  // CD, WT bits

#define ONEMEG 0x100000UL
#define TWOMEGS 0x200000UL

static char *const zpage = (char *)0;  // Page in which we live
static char *const wpage = (char *)TWOMEGS;  // The other page we map

static void
activate_paging (void)
{
  printf ("Activating paging:\n");
  uint32_t cr4 = get_cr4 ();
  printf ("cr4 = 0x%0x\n", cr4);
  cr4 |= (1<<5);
  printf ("activating PAE\n");
  set_cr4 (cr4);
  cr4 = get_cr4 ();
  printf ("cr4 = 0x%0x\n", cr4);
  set_cr3 ((uint32_t)pdpt);
  pdpt[0] = (uint64_t)(uint32_t)pd | pd_flags;  // Point to page directory
  pd[0] = 0ULL | page_flags;  // Point to first 2MB-page
  uint32_t cr0 = get_cr0 ();
  printf ("cr0 = 0x%0x\n", cr0);
  cr0 |= (1<<31);
  printf ("activating PG\n");
  set_cr0 (cr0);
  cr0 = get_cr0 ();
  printf ("cr0 = 0x%0x\n", cr0);
  printf ("Paging activated.\n");
}

static void
select_wpage (uint64_t addr)
{
  addr &= ~((uint64_t)TWOMEGS-1);
  pd[1] = addr | page_flags | cdwt_flags;  // Do not cache these pages!
  asm ("invlpg (%0)" : "=m" (*(char (*)[TWOMEGS])wpage) : "m" (pd[1]));
}

#define MAX_MMAPS 256
static int nb_mmaps;
static struct mmap_info *mmaps[MAX_MMAPS];

static void
get_mmaps (struct mb_info *info)
{
  nb_mmaps = 0;
  if ( ! ( info->flags & 0x00000040UL ) )
    return;
  printf ("Obtaining memory maps:\n");
  struct mmap_info *m = info->mmap_addr;
  while ( (char *)m < (char *)info->mmap_addr + info->mmap_length )
    {
      if ( nb_mmaps >= MAX_MMAPS )
	die ("Too many memory maps!");
      mmaps[nb_mmaps++] = m;
      printf ("mmap %d at 0x%Lx length 0x%Lx type %d\n", nb_mmaps,
	      m->base_addr, m->length, m->type);
      m = (struct mmap_info *)((char *)m+m->size+4);
    }
  printf ("Memory maps obtained.\n");
}

static char verbose;

static void
wipe_range (uint64_t start, uint64_t end)
{
  if ( end <= start )
    return;
  printf ("Will wipe from 0x%Lx to 0x%Lx\n", start, end);
  for ( uint64_t page_start=start&~((uint64_t)TWOMEGS-1) ; page_start<end ;
	page_start+=TWOMEGS )
    {
      if ( verbose )
	printf ("doing page 0x%Lx", page_start);
      select_wpage (page_start);
      uint32_t start_range = 0;
      if ( start >= page_start && start < page_start+TWOMEGS )
	start_range = start-page_start;
      uint32_t end_range = TWOMEGS;
      if ( end >= page_start && end < page_start+TWOMEGS )
	end_range = end-page_start;
      if ( verbose )
	printf (" (0x%x to 0x%x)\n", start_range, end_range);
      for ( unsigned int i=start_range/sizeof(uint32_t) ;
	    i<end_range/sizeof(uint32_t) ; i++ )
	((uint32_t *)wpage)[i] = 0xdeadbeefUL;
    }
}

#define MAX_ARGS 256
static int nb_args;
static char *args[MAX_ARGS];

static void
split_cmdline (char *cmdline)
{
  nb_args = 0;
  char new_arg = 1;
  while ( *cmdline )
    {
      if ( *cmdline == ' ' )
	{
	  *cmdline = 0;
	  new_arg = 1;
	}
      else if ( new_arg )
	{
	  args[nb_args++] = cmdline;
	  if ( nb_args >= MAX_ARGS )
	    return;
	  new_arg = 0;
	}
      cmdline++;
    }
}

static int
strcmp (const char *s1, const char *s2)
{
  while ( *s1 && *s2 )
    {
      if ( *s1 < *s2 )
	return -1;
      else if ( *s1 > *s2 )
	return 1;
      else
	{
	  s1++;
	  s2++;
	}
    }
  return 0;
}

static char
has_arg (const char *str)
{
  for ( int i=0 ; i<nb_args ; i++ )
    if ( strcmp (args[i], str) == 0 )
      return 1;
  return 0;
}

extern void _start;
extern void _end;

static const uint32_t start_addr = (uint32_t)&_start;
static const uint32_t end_addr = (uint32_t)&_end;

void
boot_main (uint32_t magic, struct mb_info *info)
{
  select_attr (CLEAR_ATTR);
  clear_screen ();
  select_attr (TEXT_ATTR);
  printf ("Booted successfully.\n");
  split_cmdline ((char *)info->cmdline);
  printf ("We are loaded from 0x%x to 0x%x\n", start_addr, end_addr);
  if ( end_addr >= TWOMEGS )
    die ("Violated assumption: loaded outside of first bigpage");
  get_mmaps (info);
  if ( nb_mmaps == 0 )
    die ("No memory map provided");
  activate_paging ();
  printf ("Now wiping:\n");
  if ( has_arg ("verbose") )
    verbose = 1;
  for ( int i=0 ; i<nb_mmaps ; i++ )
    {
      if ( mmaps[i]->type != 1 )
	continue;
      uint64_t start_region = mmaps[i]->base_addr;
      uint64_t end_region = mmaps[i]->base_addr + mmaps[i]->length;
      if ( start_region < ONEMEG )
	start_region = ONEMEG;
      if ( end_region < ONEMEG )
	end_region = ONEMEG;
      char b = 0;
      if ( start_addr >= start_region && start_addr < end_region )
	{
	  wipe_range (start_region, start_addr);
	  b = 1;
	}
      if ( end_addr >= start_region && end_addr < end_region )
	{
	  wipe_range (end_addr, end_region);
	  b = 1;
	}
      if ( ! b )
	wipe_range (start_region, end_region);
    }
  printf ("Wiping done.\n");
  if ( ! has_arg ("noreboot") )
    reboot ();
}
